/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

//
// cl_cgapi.c
//

#include "cl_local.h"

cgExport_t		*cge;

/*
=============================================================================

	CGAME EXPORT FUNCTIONS

=============================================================================
*/

/*
===============
CL_CGModule_LoadMap
===============
*/
static inline void KeepAlive (void)
{
	Netchan_Transmit (&cls.netChan, 0, NULL);
}
void CL_CGModule_LoadMap (void)
{
	int		prepInitTime;

	if (!cge)
		Com_Error (ERR_FATAL, "CL_CGModule_LoadMap: CGame not initialized!");

	cls.mapLoading = qTrue;
	cls.mapLoaded = qFalse;

	prepInitTime = Sys_Milliseconds ();

	R_BeginRegistration ();
	Snd_BeginRegistration ();

	R_GetGLConfig (&cls.glConfig);

	CL_InitImageMedia ();
	CL_InitSoundMedia ();
	SCR_UpdateScreen ();

	// Init cgame and 'keep alive' to not timeout
	KeepAlive ();
	cge->LoadMap (cl.playerNum, cls.serverProtocol, cl.attractLoop, cl.strafeHack, &cls.glConfig);
	KeepAlive ();

	// Touch before registration ends
	R_MediaInit ();

	// Start the cd track
	CDAudio_Play (atoi (cl.configStrings[CS_CDTRACK]), qTrue);

	// The subsystems can now free unneeded stuff
	R_EndRegistration ();
	Snd_EndRegistration ();

	// Clear notify lines
	Con_ClearNotify ();

	// Update the screen
	SCR_UpdateScreen ();

	// Pump message loop
	Sys_SendKeyEvents ();

	// All done!
	cls.refreshPrepped = qTrue;
	cls.mapLoading = qFalse;
	cls.mapLoaded = qTrue;
	Com_Printf (0, 0, "Map loaded in: %.2fs\n", (Sys_Milliseconds () - prepInitTime)*0.001f);
}


/*
===============
CL_CGModule_UpdateConnectInfo
===============
*/
void CL_CGModule_UpdateConnectInfo (void)
{
	if (cge) {

		// Update download info
		if (cls.download.file) {
			float	size = (float)ftell (cls.download.file);

			cge->UpdateConnectInfo (cls.serverName, cls.connectCount, cls.download.name, cls.download.percent, size);
		}
		else {
			cge->UpdateConnectInfo (cls.serverName, cls.connectCount, NULL, -1, -1);
		}
	}
}


/*
===============
CL_CGModule_BeginFrameSequence
===============
*/
void CL_CGModule_BeginFrameSequence (void)
{
	if (cge)
		cge->BeginFrameSequence (cl.frame);
}


/*
===============
CL_CGModule_NewPacketEntityState
===============
*/
void CL_CGModule_NewPacketEntityState (int entNum, entityState_t state)
{
	if (cge)
		cge->NewPacketEntityState (entNum, state);
}


/*
===============
CL_CGModule_EndFrameSequence
===============
*/
void CL_CGModule_EndFrameSequence (void)
{
	if (cge)
		cge->EndFrameSequence (cl.frame.numEntities);
}


/*
===============
CL_CGModule_GetEntitySoundOrigin
===============
*/
void CL_CGModule_GetEntitySoundOrigin (int entNum, vec3_t origin)
{
	if (cge)
		cge->GetEntitySoundOrigin (entNum, origin);
}


/*
===============
CL_CGModule_ParseConfigString
===============
*/
void CL_CGModule_ParseConfigString (int num, char *str)
{
	if (cge)
		cge->ParseConfigString (num, str);
}


/*
===============
CL_CGModule_DebugGraph
===============
*/
void CL_CGModule_DebugGraph (float value, int color)
{
	if (cge)
		cge->DebugGraph (value, color);
}


/*
===============
CL_CGModule_StartServerMessage
===============
*/
void CL_CGModule_StartServerMessage (void)
{
	if (cge)
		cge->StartServerMessage ();
}


/*
===============
CL_CGModule_ParseServerMessage
===============
*/
qBool CL_CGModule_ParseServerMessage (int command)
{
	if (cge)
		return cge->ParseServerMessage (command);

	return qFalse;
}


/*
===============
CL_CGModule_EndServerMessage
===============
*/
void CL_CGModule_EndServerMessage (void)
{
	if (cge)
		cge->EndServerMessage (cls.realTime);
}


/*
===============
CL_CGModule_Pmove
===============
*/
qBool CL_CGModule_Pmove (pMoveNew_t *pMove, float airAcceleration)
{
	if (cge) {
		cge->Pmove (pMove, airAcceleration);
		return qTrue;
	}

	return qFalse;
}


/*
===============
CL_CGModule_RegisterSounds
===============
*/
void CL_CGModule_RegisterSounds (void)
{
	if (cge)
		cge->RegisterSounds ();
}


/*
===============
CL_CGModule_RenderView
===============
*/
void CL_CGModule_RenderView (float stereoSeparation)
{
	if (cge)
		cge->RenderView (cls.realTime, cls.trueNetFrameTime, cls.trueRefreshFrameTime, stereoSeparation, cls.refreshPrepped);
}


/*
===============
CL_CGModule_SetGLConfig
===============
*/
void CL_CGModule_SetGLConfig (void)
{
	if (cge)
		cge->SetGLConfig (&cls.glConfig);
}


/*
===============
CL_CGModule_MainMenu
===============
*/
void CL_CGModule_MainMenu (void)
{
	if (cge)
		cge->MainMenu ();
}


/*
===============
CL_CGModule_ForceMenuOff
===============
*/
void CL_CGModule_ForceMenuOff (void)
{
	if (cge)
		cge->ForceMenuOff ();
}


/*
===============
CL_CGModule_MoveMouse
===============
*/
void CL_CGModule_MoveMouse (float mx, float my)
{
	if (cge)
		cge->MoveMouse (mx, my);
}


/*
===============
CL_CGModule_KeyEvent
===============
*/
void CL_CGModule_KeyEvent (int keyNum, qBool isDown)
{
	if (cge)
		cge->KeyEvent (keyNum, isDown);
}


/*
===============
CL_CGModule_ParseServerInfo
===============
*/
qBool CL_CGModule_ParseServerInfo (char *adr, char *info)
{
	if (cge)
		return cge->ParseServerInfo (adr, info);
	return qFalse;
}


/*
===============
CL_CGModule_ParseServerStatus
===============
*/
qBool CL_CGModule_ParseServerStatus (char *adr, char *info)
{
	if (cge)
		return cge->ParseServerStatus (adr, info);
	return qFalse;
}

/*
=============================================================================

	CGAME IMPORT API

=============================================================================
*/

/*
===============
CGI_NET_GetCurrentUserCmdNum
===============
*/
static int CGI_NET_GetCurrentUserCmdNum (void)
{
	return cl.cmdNum;
}


/*
===============
CGI_NET_GetPacketDropCount
===============
*/
static int CGI_NET_GetPacketDropCount (void)
{
	return cls.netChan.dropped;
}


/*
===============
CGI_NET_GetRateDropCount
===============
*/
static int CGI_NET_GetRateDropCount (void)
{
	return cl.surpressCount;
}


/*
===============
CGI_NET_GetSequenceState
===============
*/
static void CGI_NET_GetSequenceState (int *outgoingSequence, int *incomingAcknowledged)
{
	if (outgoingSequence)
		*outgoingSequence = cls.netChan.outgoingSequence;
	if (incomingAcknowledged)
		*incomingAcknowledged = cls.netChan.incomingAcknowledged;
}


/*
===============
CGI_NET_GetUserCmd
===============
*/
static void CGI_NET_GetUserCmd (int frame, userCmd_t *cmd)
{
	if (cmd)
		*cmd = cl.cmds[frame & CMD_MASK];
}


/*
===============
CGI_NET_GetUserCmdTime
===============
*/
static int CGI_NET_GetUserCmdTime (int frame)
{
	return cl.cmdTime[frame & CMD_MASK];
}

// ==========================================================================

/*
===============
CGI_GetConfigString
===============
*/
static void CGI_GetConfigString (int i, char *str, int size)
{
	if (i < 0 || i >= MAX_CONFIGSTRINGS)
		Com_Printf (0, DEV_CLIENT, S_COLOR_RED "CGI_GetConfigString: i > MAX_CONFIGSTRINGS");
	if (!str || size <= 0)
		Com_Printf (0, DEV_CLIENT, S_COLOR_RED "CGI_GetConfigString: NULL string");

	strncpy (str, cl.configStrings[i], size);
}

// ==========================================================================

static int CGI_MSG_ReadChar (void)			{ return MSG_ReadChar (&netMessage); }
static int CGI_MSG_ReadByte (void)			{ return MSG_ReadByte (&netMessage); }
static int CGI_MSG_ReadShort (void)			{ return MSG_ReadShort (&netMessage); }
static int CGI_MSG_ReadLong (void)			{ return MSG_ReadLong (&netMessage); }
static float CGI_MSG_ReadFloat (void)		{ return MSG_ReadFloat (&netMessage); }
static void CGI_MSG_ReadDir (vec3_t dir)	{ MSG_ReadDir (&netMessage, dir); }
static void CGI_MSG_ReadPos (vec3_t pos)	{ MSG_ReadPos (&netMessage, pos); }
static char *CGI_MSG_ReadString (void)		{ return MSG_ReadString (&netMessage); }

// ==========================================================================

/*
===============
CGI_R_RenderScene
===============
*/
static void CGI_R_RenderScene (refDef_t *rd)
{
	cl.refDef = *rd;
	R_RenderScene (rd);
}

// ==========================================================================

/*
===============
CGI_Free
===============
*/
static void CGI_Free (const void *ptr, const char *fileName, const int fileLine)
{
	_Mem_Free (ptr, fileName, fileLine);
}


/*
===============
CGI_Alloc
===============
*/
static void *CGI_Alloc (size_t size, qBool zeroFill, const char *fileName, const int fileLine)
{
	return _Mem_Alloc (size, zeroFill, MEMPOOL_CGAMESYS, 0, fileName, fileLine);
}


/*
===============
CGI_StrDup
===============
*/
static char *CGI_StrDup (const char *in, const char *fileName, const int fileLine)
{
	return _Mem_PoolStrDup (in, MEMPOOL_CGAMESYS, 0, fileName, fileLine);
}

// ==========================================================================

/*
===============
CL_CGameAPI_Init
===============
*/
void CL_CGameAPI_Init (void)
{
	cgImport_t	cgi;

	CL_CGameAPI_Shutdown ();	// Shutdown if already active

	Com_Printf (0, DEV_CLIENT, "\nCGame Initialization start\n");

	// Initialize pointers
	cgi.Cbuf_AddText				= Cbuf_AddText;
	cgi.Cbuf_Execute				= Cbuf_Execute;
	cgi.Cbuf_ExecuteString			= Cmd_ExecuteString;
	cgi.Cbuf_InsertText				= Cbuf_InsertText;

	cgi.CL_ResetServerCount			= CL_ResetServerCount;

	cgi.CM_BoxTrace					= CM_BoxTrace;
	cgi.CM_HeadnodeForBox			= CM_HeadnodeForBox;
	cgi.CM_InlineModel				= CM_InlineModel;
	cgi.CM_InlineModelBounds		= CM_InlineModelBounds;
	cgi.CM_InlineModelHeadNode		= CM_InlineModelHeadNode;
	cgi.CM_PointContents			= CM_PointContents;
	cgi.CM_Trace					= CM_Trace;
	cgi.CM_TransformedBoxTrace		= CM_TransformedBoxTrace;
	cgi.CM_TransformedPointContents = CM_TransformedPointContents;

	cgi.Cmd_AddCommand				= Cmd_AddCommand;
	cgi.Cmd_RemoveCommand			= Cmd_RemoveCommand;
	cgi.Cmd_Argc					= Cmd_Argc;
	cgi.Cmd_Args					= Cmd_Args;
	cgi.Cmd_Argv					= Cmd_Argv;

	cgi.Com_Error					= Com_Error;
	cgi.Com_Printf					= Com_Printf;
	cgi.Com_ClientState				= Com_ClientState;
	cgi.Com_ServerState				= Com_ServerState;

	cgi.Cvar_Register				= Cvar_Register;

	cgi.Cvar_GetIntegerValue		= Cvar_GetIntegerValue;
	cgi.Cvar_GetStringValue			= Cvar_GetStringValue;
	cgi.Cvar_GetFloatValue			= Cvar_GetFloatValue;

	cgi.Cvar_VariableSet			= Cvar_VariableSet;
	cgi.Cvar_VariableSetValue		= Cvar_VariableSetValue;
	cgi.Cvar_VariableReset			= Cvar_VariableReset;

	cgi.Cvar_Set					= Cvar_Set;
	cgi.Cvar_SetValue				= Cvar_SetValue;
	cgi.Cvar_Reset					= Cvar_Reset;

	cgi.FS_CreatePath				= FS_CreatePath;
	cgi.FS_FreeFileList				= _FS_FreeFileList;
	cgi.FS_Gamedir					= FS_Gamedir;
	cgi.FS_FindFiles				= FS_FindFiles;
	cgi.FS_LoadFile					= FS_LoadFile;
	cgi.FS_FreeFile					= _FS_FreeFile;
	cgi.FS_NextPath					= FS_NextPath;
	cgi.FS_Read						= FS_Read;
	cgi.FS_Write					= FS_Write;
	cgi.FS_Seek						= FS_Seek;
	cgi.FS_OpenFile					= FS_OpenFile;
	cgi.FS_CloseFile				= FS_CloseFile;

	cgi.GetConfigString				= CGI_GetConfigString;

	cgi.Key_ClearStates				= Key_ClearStates;
	cgi.Key_GetKeyDest				= Key_GetKeyDest;
	cgi.Key_GetBindingBuf			= Key_GetBindingBuf;
	cgi.Key_IsDown					= Key_IsDown;
	cgi.Key_KeynumToString			= Key_KeynumToString;
	cgi.Key_SetBinding				= Key_SetBinding;
	cgi.Key_SetKeyDest				= Key_SetKeyDest;
	cgi.Key_InsertOn				= Key_InsertOn;
	cgi.Key_CapslockOn				= Key_CapslockOn;
	cgi.Key_ShiftDown				= Key_ShiftDown;

	cgi.Mem_Alloc					= CGI_Alloc;
	cgi.Mem_Free					= CGI_Free;
	cgi.Mem_StrDup					= CGI_StrDup;

	cgi.MSG_ReadChar				= CGI_MSG_ReadChar;
	cgi.MSG_ReadByte				= CGI_MSG_ReadByte;
	cgi.MSG_ReadShort				= CGI_MSG_ReadShort;
	cgi.MSG_ReadLong				= CGI_MSG_ReadLong;
	cgi.MSG_ReadFloat				= CGI_MSG_ReadFloat;
	cgi.MSG_ReadDir					= CGI_MSG_ReadDir;
	cgi.MSG_ReadPos					= CGI_MSG_ReadPos;
	cgi.MSG_ReadString				= CGI_MSG_ReadString;

	cgi.NET_GetCurrentUserCmdNum	= CGI_NET_GetCurrentUserCmdNum;
	cgi.NET_GetPacketDropCount		= CGI_NET_GetPacketDropCount;
	cgi.NET_GetRateDropCount		= CGI_NET_GetRateDropCount;
	cgi.NET_GetSequenceState		= CGI_NET_GetSequenceState;
	cgi.NET_GetUserCmd				= CGI_NET_GetUserCmd;
	cgi.NET_GetUserCmdTime			= CGI_NET_GetUserCmdTime;

	cgi.R_AddEntity					= R_AddEntity;
	cgi.R_AddPoly					= R_AddPoly;
	cgi.R_AddLight					= R_AddLight;
	cgi.R_AddLightStyle				= R_AddLightStyle;

	cgi.R_ClearScene				= R_ClearScene;

	cgi.R_CullBox					= R_CullBox;
	cgi.R_CullSphere				= R_CullSphere;
	cgi.R_CullVisNode				= R_CullVisNode;

	cgi.R_DrawFill					= R_DrawFill;
	cgi.R_DrawPic					= R_DrawPic;

	cgi.R_GetClippedFragments		= R_GetClippedFragments;
	cgi.R_GetGLConfig				= R_GetGLConfig;
	cgi.R_GetImageSize				= R_GetImageSize;

	cgi.R_RegisterMap				= R_RegisterMap;
	cgi.R_RegisterModel				= R_RegisterModel;
	cgi.R_ModelBounds				= R_ModelBounds;

	cgi.R_UpdateScreen				= SCR_UpdateScreen;
	cgi.R_RenderScene				= CGI_R_RenderScene;
	cgi.R_BeginFrame				= R_BeginFrame;
	cgi.R_EndFrame					= R_EndFrame;

	cgi.R_LightPoint				= R_LightPoint;
	cgi.R_TransformVectorToScreen	= R_TransformVectorToScreen;
	cgi.R_SetSky					= R_SetSky;

	cgi.R_RegisterPic				= R_RegisterPic;
	cgi.R_RegisterPoly				= R_RegisterPoly;
	cgi.R_RegisterSkin				= R_RegisterSkin;
	cgi.R_RegisterTexture			= R_RegisterTexture;

	cgi.Snd_RegisterSound			= Snd_RegisterSound;
	cgi.Snd_StartLocalSound			= Snd_StartLocalSound;
	cgi.Snd_StartSound				= Snd_StartSound;
	cgi.Snd_Update					= Snd_Update;

	cgi.Sys_FindClose				= Sys_FindClose;
	cgi.Sys_FindFirst				= Sys_FindFirst;
	cgi.Sys_GetClipboardData		= Sys_GetClipboardData;
	cgi.Sys_Milliseconds			= Sys_Milliseconds;
	cgi.Sys_SendKeyEvents			= Sys_SendKeyEvents;

	// Get the cgame api
	cge = (cgExport_t *) Sys_LoadLibrary (LIB_CGAME, &cgi);
	if (!cge) {
		cls.moduleCGameActive = qFalse;
		Com_Error (ERR_FATAL, "Find/load of CGame library failed!\n");
	}

	// Checking the cg api version
	if (cge->apiVersion != CGAME_APIVERSION) {
		Com_Error (ERR_FATAL, "CL_CGameAPI_Init: incompatible apiVersion (%i != %i)\n", cge->apiVersion, CGAME_APIVERSION);

		Sys_UnloadLibrary (LIB_CGAME);
		cge = NULL;
		cls.moduleCGameActive = qFalse;
	}

	cge->Init ();

	cls.moduleCGameActive = qTrue;
}


/*
===============
CL_CGameAPI_Shutdown
===============
*/
void CL_CGameAPI_Shutdown (void)
{
	if (!cge)
		return;

	Com_Printf (0, DEV_CLIENT, "CGame Shutdown start\n");

	// Tell the module to shutdown locally and unload dll
	cge->Shutdown ();
	Sys_UnloadLibrary (LIB_CGAME);
	cls.mapLoaded = qFalse;
	cls.moduleCGameActive = qFalse;
	cge = NULL;

	// Free the memory pool
	Mem_FreePool (MEMPOOL_CGAMESYS);
}
