/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

//
// common.h
// Definitions common between client and server, but not game.dll
//

#ifndef __COMMON_H__
#define __COMMON_H__

#include "../shared/shared.h"
#include "files.h"
#include "font.h"
#include "protocol.h"
#include "cm_public.h"
#include "alias.h"
#include "cmd.h"
#include "cvar.h"
#include "memory.h"

#define Q2_VERSION		3.21			// Quake2 version
#define EGL_VERSTR		"0.2.1"

#define BASE_MODDIRNAME		"baseq2"

/*
=============================================================================

	MISCELLANEOUS

=============================================================================
*/

extern cVar_t	*developer;
extern cVar_t	*dedicated;

// hash optimizing
uInt	Com_HashFileName (const char *fileName, const int hashSize);
uInt	Com_HashGeneric (const char *name, const int hashSize);

// client/server interactions
void	Com_BeginRedirect (int target, char *buffer, int bufferSize, void (*flush));
void	Com_EndRedirect (void);
void	Com_Quit (void);

// client/server states
int		Com_ClientState (void);
void	Com_SetClientState (int state);

enum {
	SS_DEAD,			// no map loaded
	SS_LOADING,			// spawning level edicts
	SS_GAME,			// actively running
	SS_CINEMATIC,		// playing a cinematic
	SS_DEMO,			// playing a demo
	SS_PIC				// just showing a pic
};
int		Com_ServerState (void);		// this should have just been a cvar...
void	Com_SetServerState (int state);

// initialization and processing
void	Com_Init (int argc, char **argv);
void	__fastcall Com_Frame (int msec);
void	Com_Shutdown (void);

// crc and checksum
byte	Com_BlockSequenceCRCByte (byte *base, int length, int sequence);
uInt	Com_BlockChecksum (void *buffer, int length);

/*
==============================================================================

	NON-PORTABLE SYSTEM SERVICES

==============================================================================
*/

enum {
	LIB_CGAME,
	LIB_GAME
};

extern	int	sys_currTime;		// time returned by last Sys_Milliseconds
int		Sys_Milliseconds (void);

void	Sys_Init (void);
void	Sys_AppActivate (void);

void	Sys_UnloadLibrary (int libType);
void	*Sys_LoadLibrary (int libType, void *parms);

void	Sys_SendKeyEvents (void);
void	Sys_Error (char *error, ...);
void	Sys_Print (const char *message);
void	Sys_Quit (void);
char	*Sys_GetClipboardData (void);

void	Sys_Mkdir (char *path);

// pass in an attribute mask of things you wish to REJECT
char	*Sys_FindFirst (char *path, uInt mustHave, uInt cantHave);
char	*Sys_FindNext (uInt mustHave, uInt cantHave);
void	Sys_FindClose (void);

int		Sys_FindFiles (char *path, char *pattern, char **fileList, int maxFiles, int fileCount, qBool recurse, qBool addFiles, qBool addDirs);

// ==========================================================================

void	Sys_SetErrorText (const char *buf);
void	Conbuf_AppendText (const char *pMsg);
char	*Sys_ConsoleInput (void);
void	Sys_ShowConsole (int visLevel, qBool quitOnClose);
void	Sys_DestroyConsole (void);
void	Sys_CreateConsole (void);

/*
=============================================================================

	CLIENT / SERVER SYSTEMS

=============================================================================
*/

//
// console.c
//

void	Con_Init (void);
void	Con_Print (short flags, const char *text);

//
// cl_cgapi.c
//

qBool	CL_CGModule_Pmove (pMoveNew_t *pMove, float airAcceleration);

//
// cl_main.c
//

void	CL_ForcePacket (void);

void	CL_Disconnect (qBool openMenu);
void	__fastcall CL_Frame (int msec);
void	CL_Init (void);
void	CL_Shutdown (qBool error);

//
// cl_screen.c
//

void	SCR_BeginLoadingPlaque (void);
void	SCR_EndLoadingPlaque (void);

//
// keys.c
//

void	Key_WriteBindings (FILE *f);

char	*Key_GetBindingBuf (int binding);
qBool	Key_IsDown (int keyNum);

void	Key_Init (void);

//
// sv_main.c
//

void	SV_Init (void);
void	SV_Shutdown (char *finalmsg, qBool reconnect, qBool crashing);
void	SV_Frame (int msec);

#endif // __COMMON_H__
