/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "x11.h"
#include "../client/cl_local.h"

extern qBool	mlooking;
extern cVar_t *m_filter;
extern cVar_t *in_mouse;
extern cVar_t *in_dgamouse;
extern cVar_t *freelook;

extern qBool mouse_avail;
extern int mouse_buttonstate;
extern int mouse_oldbuttonstate;
extern int mx, my;
extern int old_mouse_x, old_mouse_y;

extern x11display_t x11display;

qBool mouse_grabbed = qFalse;

void Force_CenterView_f (void)
{
	cl.viewAngles[PITCH] = 0;
}

void IN_Restart_f (void)
{
}

void IN_Commands (void)
{
}

void IN_Move (userCmd_t *cmd)
{
	int kd = Key_GetKeyDest ();

	if (!mouse_avail) 
		return;

	HandleEvents(); //Make sure mouse state is up to date

	if (kd == KD_MENU || kd == KD_CONSOLE) {
		if (mouse_grabbed) X11_SetMouseGrab(qFalse);

		if (kd == KD_MENU)
			CL_CGModule_MoveMouse (mx, my);
		mx = my = 0;
		return;
	}
	if (!mouse_grabbed) X11_SetMouseGrab(qTrue);

	if (m_filter->integer)
	{
		mx = (mx + old_mouse_x) * 0.5;
		my = (my + old_mouse_y) * 0.5;
	}
  
	old_mouse_x = mx;
	old_mouse_y = my;
  
	mx *= sensitivity->value;
	my *= sensitivity->value;
  
	cl.viewAngles[YAW] -= m_yaw->value * mx;
	cl.viewAngles[PITCH] += m_pitch->value * my;
    
	mx = my = 0;
}

void IN_Activate (qBool active)
{
  int kd = Key_GetKeyDest();
  if (active)
    X11_SetMouseGrab(kd != KD_MENU && kd != KD_CONSOLE);
  else
    X11_SetMouseGrab(qFalse);
}


void IN_Init(void)
{
	// mouse variables
	m_filter = Cvar_Register ("m_filter", "0", 0);
	in_mouse = Cvar_Register ("in_mouse", "1", CVAR_ARCHIVE);
	in_dgamouse = Cvar_Register ("in_dgamouse", "1", CVAR_ARCHIVE);
	freelook = Cvar_Register ("cl_freelook", "0", 0);
	lookstrafe = Cvar_Register ("lookstrafe", "0", 0);
	sensitivity = Cvar_Register ("sensitivity", "3", 0);
	m_pitch = Cvar_Register ("m_pitch", "0.022", 0);
	m_yaw = Cvar_Register ("m_yaw", "0.022", 0);
	m_forward = Cvar_Register ("m_forward", "1", 0);
	m_side = Cvar_Register ("m_side", "0.8", 0);

	Cmd_AddCommand ("force_centerview", Force_CenterView_f, "Force the screen to a center view");

	mx = my = 0.0;
	mouse_avail = qTrue;
}

void IN_Shutdown (void)
{
	if (mouse_avail)
	{
		mouse_avail = qFalse;
		Cmd_RemoveCommand ("force_centerview", NULL);
	}
}

void IN_Frame (void)
{
//	if ( cls.key_dest == key_console || cls.key_dest == key_menu)
//		IN_Activate (qFalse);
//	else
		IN_Activate (qTrue);

	HandleEvents();
}


