/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

//
// r_glstate.c
// OpenGL state handling and setup
//

#include "r_local.h"

glState_t	glState;

/*
===============
GL_Enable
===============
*/
void GL_Enable (GLenum cap)
{
	switch (cap) {
	case GL_ALPHA_TEST:
		if (glState.alphaTestOn)
			return;
		glState.alphaTestOn = qTrue;
		break;

	case GL_BLEND:
		if (glState.blendingOn)
			return;
		glState.blendingOn = qTrue;
		break;

	case GL_CULL_FACE:
		if (glState.cullingOn)
			return;
		glState.cullingOn = qTrue;
		break;

	case GL_DEPTH_TEST:
		if (glState.depthTestOn)
			return;
		glState.depthTestOn = qTrue;
		break;

	case GL_POLYGON_OFFSET_FILL:
		if (glState.polyOffsetOn)
			return;
		glState.polyOffsetOn = qTrue;
		break;

	case GL_STENCIL_TEST:
		if (glState.stencilTestOn)
			return;
		if (!glStatic.useStencil)
			return;
		glState.stencilTestOn = qTrue;
		break;

	case GL_TEXTURE_1D:
		if (glState.tex1DOn[glState.texUnit])
			return;
		glState.tex1DOn[glState.texUnit] = qTrue;
		break;

	case GL_TEXTURE_2D:
		if (glState.tex2DOn[glState.texUnit])
			return;
		glState.tex2DOn[glState.texUnit] = qTrue;
		break;

	case GL_TEXTURE_3D:
		if (glState.tex3DOn[glState.texUnit])
			return;
		if (!glConfig.extTex3D)
			return;
		glState.tex3DOn[glState.texUnit] = qTrue;
		break;

	case GL_TEXTURE_CUBE_MAP_ARB:
		if (glState.texCMOn[glState.texUnit])
			return;
		if (!glConfig.extTexCubeMap)
			return;
		glState.texCMOn[glState.texUnit] = qTrue;
		break;

	case GL_TEXTURE_GEN_S:
		if (glState.texGenSOn[glState.texUnit])
			return;
		glState.texGenSOn[glState.texUnit] = qTrue;
		break;

	case GL_TEXTURE_GEN_T:
		if (glState.texGenTOn[glState.texUnit])
			return;
		glState.texGenTOn[glState.texUnit] = qTrue;
		break;

	case GL_TEXTURE_GEN_R:
		if (glState.texGenROn[glState.texUnit])
			return;
		glState.texGenROn[glState.texUnit] = qTrue;
		break;

	case GL_TEXTURE_GEN_Q:
		if (glState.texGenQOn[glState.texUnit])
			return;
		glState.texGenQOn[glState.texUnit] = qTrue;
		break;

	case GL_FRAGMENT_PROGRAM_ARB:
		if (glState.fragProgOn)
			return;
		if (!glConfig.extFragmentProgram)
			return;
		glState.fragProgOn = qTrue;
		break;

	case GL_VERTEX_PROGRAM_ARB:
		if (glState.vertProgOn)
			return;
		if (!glConfig.extVertexProgram)
			return;
		glState.vertProgOn = qTrue;
		break;
	}

	qglEnable (cap);
}


/*
===============
GL_Disable
===============
*/
void GL_Disable (GLenum cap)
{
	switch (cap) {
	case GL_ALPHA_TEST:
		if (!glState.alphaTestOn)
			return;
		glState.alphaTestOn = qFalse;
		break;

	case GL_BLEND:
		if (!glState.blendingOn)
			return;
		glState.blendingOn = qFalse;
		break;

	case GL_CULL_FACE:
		if (!glState.cullingOn)
			return;
		glState.cullingOn = qFalse;
		break;

	case GL_DEPTH_TEST:
		if (!glState.depthTestOn)
			return;
		glState.depthTestOn = qFalse;
		break;

	case GL_POLYGON_OFFSET_FILL:
		if (!glState.polyOffsetOn)
			return;
		glState.polyOffsetOn = qFalse;
		break;

	case GL_STENCIL_TEST:
		if (!glState.stencilTestOn)
			return;
		if (!glStatic.useStencil)
			return;
		glState.stencilTestOn = qFalse;
		break;

	case GL_TEXTURE_1D:
		if (!glState.tex1DOn[glState.texUnit])
			return;
		glState.tex1DOn[glState.texUnit] = qFalse;
		break;

	case GL_TEXTURE_2D:
		if (!glState.tex2DOn[glState.texUnit])
			return;
		glState.tex2DOn[glState.texUnit] = qFalse;
		break;

	case GL_TEXTURE_3D:
		if (!glState.tex3DOn[glState.texUnit])
			return;
		if (!glConfig.extTex3D)
			return;
		glState.tex3DOn[glState.texUnit] = qFalse;
		break;

	case GL_TEXTURE_CUBE_MAP_ARB:
		if (!glState.texCMOn[glState.texUnit])
			return;
		if (!glConfig.extTexCubeMap)
			return;
		glState.texCMOn[glState.texUnit] = qFalse;
		break;

	case GL_TEXTURE_GEN_S:
		if (!glState.texGenSOn[glState.texUnit])
			return;
		glState.texGenSOn[glState.texUnit] = qFalse;
		break;

	case GL_TEXTURE_GEN_T:
		if (!glState.texGenTOn[glState.texUnit])
			return;
		glState.texGenTOn[glState.texUnit] = qFalse;
		break;

	case GL_TEXTURE_GEN_R:
		if (!glState.texGenROn[glState.texUnit])
			return;
		glState.texGenROn[glState.texUnit] = qFalse;
		break;

	case GL_TEXTURE_GEN_Q:
		if (!glState.texGenQOn[glState.texUnit])
			return;
		glState.texGenQOn[glState.texUnit] = qFalse;
		break;

	case GL_FRAGMENT_PROGRAM_ARB:
		if (!glState.fragProgOn)
			return;
		if (!glConfig.extFragmentProgram)
			return;
		glState.fragProgOn = qFalse;
		break;

	case GL_VERTEX_PROGRAM_ARB:
		if (!glState.vertProgOn)
			return;
		if (!glConfig.extVertexProgram)
			return;
		glState.vertProgOn = qFalse;
		break;
	}

	qglDisable (cap);
}


/*
===============
GL_BlendFunc
===============
*/
void GL_BlendFunc (GLenum sFactor, GLenum dFactor)
{
	if (glState.blendSFactor != sFactor || glState.blendDFactor != dFactor) {
		qglBlendFunc (sFactor, dFactor);
		glState.blendSFactor = sFactor;
		glState.blendDFactor = dFactor;
	}
}


/*
===============
GL_CullFace
===============
*/
void GL_CullFace (GLenum mode)
{
	if (glState.cullFace != mode) {
		qglCullFace (mode);
		glState.cullFace = mode;
	}
}


/*
===============
GL_DepthFunc
===============
*/
void GL_DepthFunc (GLenum func)
{
	if (glState.depthFunc != func) {
		qglDepthFunc (func);
		glState.depthFunc = func;
	}
}


/*
===============
GL_ShadeModel
===============
*/
void GL_ShadeModel (GLenum mode)
{
	if (glState.shadeModel != mode) {
		qglShadeModel (mode);
		glState.shadeModel = mode;
	}
}


/*
==================
GL_SetDefaultState

Sets our default OpenGL state
==================
*/
void GL_SetDefaultState (void)
{
	int		i;

	qglFinish ();

	qglColor4f (1, 1, 1, 1);
	qglClearColor (0.5f, 0.5f, 0.5f, 1.0f);

	qglEnable (GL_SCISSOR_TEST);

	qglPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
	qglPolygonOffset (0, 0);

	// Texture-unit specific
	for (i=MAX_TEXUNITS-1 ; i>=0 ; i--) {
		glState.tex1DOn[i] = qFalse;
		glState.tex1DBound[i] = 0;
		glState.tex2DBound[i] = 0;
		glState.tex3DOn[i] = qFalse;
		glState.tex3DBound[i]  = 0;
		glState.texCMOn[i] = qFalse;
		glState.texCMBound[i] = 0;

		glState.texEnvModes[i] = 0;
		glState.texGenSOn[i] = qFalse;
		glState.texGenTOn[i] = qFalse;
		glState.texGenROn[i] = qFalse;
		glState.texGenQOn[i] = qFalse;
		if (i >= glConfig.maxTexUnits) {
			glState.tex2DOn[i] = qFalse;
			continue;
		}

		// Texture
		GL_SelectTexture (i);
		qglDisable (GL_TEXTURE_1D);
		if (glConfig.extTex3D)
			qglDisable (GL_TEXTURE_3D);

		if (i == 0) {
			qglEnable (GL_TEXTURE_2D);
			glState.tex2DOn[i] = qTrue;
		}
		else {
			qglDisable (GL_TEXTURE_2D);
			glState.tex2DOn[i] = qFalse;
		}

		// Cube mapping
		if (glConfig.extTexCubeMap)
			qglDisable (GL_TEXTURE_CUBE_MAP_ARB);

		// Texgen
		qglDisable (GL_TEXTURE_GEN_S);
		qglDisable (GL_TEXTURE_GEN_T);
		qglDisable (GL_TEXTURE_GEN_R);
		qglDisable (GL_TEXTURE_GEN_Q);
	}

	// Fragment programs
	if (glConfig.extFragmentProgram)
		qglDisable (GL_FRAGMENT_PROGRAM_ARB);
	glState.fragProgOn = qFalse;

	// Vertex programs
	if (glConfig.extVertexProgram)
		qglDisable (GL_VERTEX_PROGRAM_ARB);
	glState.vertProgOn = qFalse;

	// Stencil testing
	qglDisable (GL_STENCIL_TEST);
	glState.stencilTestOn = qFalse;

	// Polygon offset testing
	qglDisable (GL_POLYGON_OFFSET_FILL);
	glState.polyOffsetOn = qFalse;

	// Depth testing
	qglDisable (GL_DEPTH_TEST);
	qglDepthFunc (GL_LEQUAL);
	qglDepthRange (0, 1);
	glState.depthTestOn = qFalse;
	glState.depthFunc = GL_LEQUAL;

	// Face culling
	qglDisable (GL_CULL_FACE);
	qglCullFace (GL_FRONT);
	glState.cullingOn = qFalse;
	glState.cullFace = GL_FRONT;

	// Alpha testing
	qglDisable (GL_ALPHA_TEST);
	qglAlphaFunc (GL_GREATER, 0.666f);
	glState.alphaTestOn = qFalse;

	// Blending
	qglDisable (GL_BLEND);
	qglBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glState.blendingOn = qFalse;
	glState.blendSFactor = GL_SRC_ALPHA;
	glState.blendDFactor = GL_ONE_MINUS_SRC_ALPHA;

	// Model shading
	qglShadeModel (GL_SMOOTH);
	glState.shadeModel = GL_SMOOTH;

	// Check for errors
	GL_CheckForError ("GL_SetDefaultState");
}
